package de.unikoblenz.fgbks.base.builder;

import java.io.Serializable;
import java.lang.reflect.Field;

/**
 * A abstract builder for creating java pojo classes.
 *
 * @param <E> the type of the object being constructed
 * @param <B> the instance type of the builder
 */
public abstract class AbstractBuilder<E, B extends AbstractBuilder<E, B>> implements Serializable {

  private static final String VALUE_FIELD_NAME = "value";

  protected E value;

  protected AbstractBuilder() {
    value = newInstance();
  }

  protected AbstractBuilder(E v) {
    value = v;
  }

  /**
   * Return the value as accessible {@link Field}.
   *
   * @return the value as accessible {@link Field}
   * @throws IllegalStateException if class does not have the named field
   */
  public static Field getAccessibleValueField() throws IllegalStateException {
    try {
      Field valueField = AbstractBuilder.class.getDeclaredField(VALUE_FIELD_NAME);
      valueField.setAccessible(true);
      return valueField;
    } catch (NoSuchFieldException e) {
      throw new IllegalStateException(
          String.format(
              "%s must have field with name '%s'",
              AbstractBuilder.class.getSimpleName(), VALUE_FIELD_NAME),
          e);
    }
  }

  /**
   * Return the current Builder.
   *
   * @return the current Builder
   */
  protected abstract B thisBuilder();

  protected abstract E newInstance();

  /**
   * Check if the fields in the value object are valid.
   */
  protected abstract void validate();

  /**
   * Start the building process and return the finished object. While building the new object the
   * function {@link DefaultBuilder#validate()} is called and should be passed without an exception.
   * Every time this function is called a new instance of the type is returned.
   *
   * @return the new created object
   */
  public E build() {
    validate();
    E result = value;
    value = newInstance();
    return result;
  }
}
