package de.unikoblenz.fgbks.base.utils.boundary;

import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreater;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheck;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreater;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.impl.InvalidBoundary;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VTypeRef;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import org.apache.commons.lang3.Validate;

/**
 * The abstract class for a boundary object, which implements the {@link Boundary} interface.
 *
 * @param <T> the comparable value type
 */
public abstract class AbstractBoundary<T extends Comparable<? super T>> implements Boundary {

  private HashMap<BoundaryCheckType, BoundaryCheck> checkFunctions;
  private HashMap<BoundaryBiCreaterType, BoundaryBiCreater> biCreaterFunctions;
  private HashMap<BoundaryCreaterType, BoundaryCreater> createrFunctions;
  private Class<T> type;
  protected String text;

  /**
   * Create a new boundary with the given type and the given text. This constructor parse the text.
   *
   * @param text the text as string, which will be parsed to the new boundary
   * @param type the type of the boundary
   */
  protected AbstractBoundary(String text, Class<T> type) {
    this.text = text;
    this.type = type;
    parse(text);
    validate();
    checkFunctions = new HashMap<>(getChecker());
    biCreaterFunctions = new HashMap<>(getBiCreater());
    createrFunctions = new HashMap<>(getCreater());
  }

  /**
   * Create a new boundary with the given type. This constructor parse the empty text value.
   *
   * @param type the type of the boundary
   */
  protected AbstractBoundary(Class<T> type) {
    this("", type);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public String getText() {
    return text;
  }

  /** {@inheritDoc} */
  @Override
  public VTypeRef getTypeRef() {
    return VTypeRef.getTypeRefFromJavaClass(type);
  }

  /** {@inheritDoc} */
  @Override
  public boolean checkWith(BoundaryCheckType checkType, Boundary other) {
    if (other instanceof InvalidBoundary) {
      return false;
    }
    BoundaryCheck checkFunction = getCheckFunction(checkType);
    if (checkFunction == null || !isValidOperation(checkType, other)) {
      throw new BoundaryOperationNotSupportedException(); // TODO: add error text
    }
    return checkFunction.check(this, other); // TODO, remove RAW type
  }

  /** {@inheritDoc} */
  @Override
  public Optional<Boundary> create(BoundaryCreaterType createrType) {
    BoundaryCreater creater = getCreaterFunction(createrType);
    if (creater == null || !isValidOperation(createrType)) {
      throw new BoundaryOperationNotSupportedException(); // TODO: add error text
    }
    return creater.create(this);
  }

  /** {@inheritDoc} */
  @Override
  public Optional<Boundary> createBi(BoundaryBiCreaterType biCreaterType, Boundary other) {
    if (other instanceof InvalidBoundary) {
      return Optional.empty();
    }
    BoundaryBiCreater biCreater = getBiCreaterFunction(biCreaterType);
    if (biCreater == null || !isValidOperation(biCreaterType, other)) {
      throw new BoundaryOperationNotSupportedException(); // TODO: add error text
    }
    return biCreater.create(this, other); // TODO, remove RAW type
  }

  /**
   * Check, if the {@link BoundaryCheckType} is a valid operation for this boundary.
   *
   * @param checkType the {@link BoundaryCheckType}
   * @param other the other bounday
   * @return true, if the operation is valid.
   */
  public boolean isValidOperation(BoundaryCheckType checkType, Boundary other) {
    return this.getTypeRef() == other.getTypeRef() && isValidOperation(checkType);
  }

  /**
   * Check, if the {@link BoundaryCheckType} is a valid operation for this boundary.
   *
   * @param checkType the {@link BoundaryCheckType}
   * @return true, if the operation is valid.
   */
  public boolean isValidOperation(BoundaryCheckType checkType) {
    return checkFunctions.containsKey(Validate.notNull(checkType));
  }

  /**
   * Check, if the {@link BoundaryBiCreaterType} is a valid operation for this boundary.
   *
   * @param combineType the {@link BoundaryBiCreaterType}
   * @param other the other bounday
   * @return true, if the operation is valid.
   */
  public boolean isValidOperation(BoundaryBiCreaterType combineType, Boundary other) {
    return this.getTypeRef() == other.getTypeRef() && isValidOperation(combineType);
  }

  /**
   * Check, if the {@link BoundaryBiCreaterType} is a valid operation for this boundary.
   *
   * @param combineType the {@link BoundaryBiCreaterType}
   * @return true, if the operation is valid.
   */
  public boolean isValidOperation(BoundaryBiCreaterType combineType) {
    return biCreaterFunctions.containsKey(Validate.notNull(combineType));
  }

  /**
   * Check, if the {@link BoundaryCreaterType} is a valid operation for this boundary.
   *
   * @param createrType the {@link BoundaryCreaterType}
   * @return true, if the operation is valid.
   */
  public boolean isValidOperation(BoundaryCreaterType createrType) {
    return createrFunctions.containsKey(Validate.notNull(createrType));
  }

  protected BoundaryCheck getCheckFunction(BoundaryCheckType checkType) {
    return checkFunctions.get(checkType);
  }

  protected BoundaryCreater getCreaterFunction(BoundaryCreaterType createrType) {
    return createrFunctions.get(createrType);
  }

  protected BoundaryBiCreater getBiCreaterFunction(BoundaryBiCreaterType biCreaterTypeType) {
    return biCreaterFunctions.get(biCreaterTypeType);
  }

  /**
   * Parse the boundary object from the given text.
   *
   * @param text the text as string
   */
  protected abstract void parse(String text);

  /**
   * Check, if the boundary is a valid boundary.
   *
   * @throws BoundaryParseException
   */
  protected abstract void validate() throws BoundaryParseException;

  /**
   * Get the map of Checker.
   *
   * @return the map of {@link BoundaryCheckType}: {@link BoundaryCheck}
   */
  protected abstract Map<BoundaryCheckType, BoundaryCheck> getChecker();

  /**
   * Get the map of biCreater.
   *
   * @return the map of {@link BoundaryBiCreaterType}: {@link BoundaryBiCreater}
   */
  protected abstract Map<BoundaryBiCreaterType, BoundaryBiCreater> getBiCreater();

  /**
   * Get the map of creater.
   *
   * @return the map of {@link BoundaryCreaterType}: {@link BoundaryCreater}
   */
  protected abstract Map<BoundaryCreaterType, BoundaryCreater> getCreater();

  @Override
  public String toString() {
    return "{" + getText() + "},{" + getParsedText() + "}";
  }
}
