package de.unikoblenz.fgbks.base.utils.boundary.bicreater;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_NOT_IN_CONTACT;

import de.unikoblenz.fgbks.base.utils.boundary.AbstractGrowingBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.impl.BooleanBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.impl.IntegerBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.impl.LongBoundary;
import java.util.Optional;

/**
 * The combine creator implementation of {@link BoundaryBiCreaterType#APPEND} for {@link
 * AbstractGrowingBoundary}. <br>
 * The result is a new boundary with the combined range of both boundaries, or an empty optional, if
 * the creation is not possible. Both boundaries are not in contact. <br>
 * In: {@code b1: [10..30[; b2: [30..50]} <br>
 * Out: {@code ]10..50[} <br>
 */
public class BiCreaterAppend extends AbstractBoundaryBiCreater<AbstractGrowingBoundary> {

  private static final BiCreaterAppend instance = new BiCreaterAppend();

  private BiCreaterAppend() {
    super();
  }

  /**
   * Get the current instance of this class.
   *
   * @return the singleton instance
   */
  public static BiCreaterAppend getInstance() {
    return instance;
  }

  @Override
  public BoundaryBiCreaterType getType() {
    return BoundaryBiCreaterType.APPEND;
  }

  @Override
  public Optional<AbstractGrowingBoundary> create(
      AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    if (b1.checkWith(IS_NOT_IN_CONTACT, b2)) {
      if (b1 instanceof BooleanBoundary) {
        return createrBoolean((BooleanBoundary) b1, (BooleanBoundary) b2);
      }
      if (b1 instanceof IntegerBoundary) {
        return createrInt(b1, b2);
      }
      if (b1 instanceof LongBoundary) {
        return createrLong(b1, b2);
      }
      if (b1.getUpperBound().equals(b2.getLowerBound())
          && b1.getUpperBoundType() != b2.getLowerBoundType()) {
        return Optional.of(
            b1.getCopy(
                b1.getLowerBound(),
                b2.getUpperBound(),
                b1.getLowerBoundType(),
                b2.getUpperBoundType()));
      } else if (b2.getUpperBound().equals(b1.getLowerBound())
          && b2.getUpperBoundType() != b1.getLowerBoundType()) {
        return Optional.of(
            b1.getCopy(
                b2.getLowerBound(),
                b1.getUpperBound(),
                b2.getLowerBoundType(),
                b1.getUpperBoundType()));
      }
    }
    return Optional.empty();
  }

  private Optional<AbstractGrowingBoundary> createrInt(
      AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    int b1L = (int) b1.getLowerBound();
    int b1U = (int) b1.getUpperBound();
    int b2L = (int) b2.getLowerBound();
    int b2U = (int) b2.getUpperBound();
    if (b1U + 1 == b2L
        && b1.getMaxValue().compareTo(b1U) != 0
        && b1.getUpperBoundType() == INCLUSIVE
        && b2.getLowerBoundType() == INCLUSIVE) {
      return Optional.of(
          b1.getCopy(
              b1.getLowerBound(),
              b2.getUpperBound(),
              b1.getLowerBoundType(),
              b2.getUpperBoundType()));
    } else if (b2U + 1 == b1L
        && b2.getMaxValue().compareTo(b2U) != 0
        && b2.getUpperBoundType() == INCLUSIVE
        && b1.getLowerBoundType() == INCLUSIVE) {
      return Optional.of(
          b1.getCopy(
              b2.getLowerBound(),
              b1.getUpperBound(),
              b2.getLowerBoundType(),
              b1.getUpperBoundType()));
    }
    return Optional.empty();
  }

  private Optional<AbstractGrowingBoundary> createrLong(
      AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    long b1L = (long) b1.getLowerBound();
    long b1U = (long) b1.getUpperBound();
    long b2L = (long) b2.getLowerBound();
    long b2U = (long) b2.getUpperBound();
    if (b1U + 1 == b2L
        && b1.getMaxValue().compareTo(b1U) != 0
        && b1.getUpperBoundType() == INCLUSIVE
        && b2.getLowerBoundType() == INCLUSIVE) {
      return Optional.of(
          b1.getCopy(
              b1.getLowerBound(),
              b2.getUpperBound(),
              b1.getLowerBoundType(),
              b2.getUpperBoundType()));
    } else if (b2U + 1 == b1L
        && b2.getMaxValue().compareTo(b2U) != 0
        && b2.getUpperBoundType() == INCLUSIVE
        && b1.getLowerBoundType() == INCLUSIVE) {
      return Optional.of(
          b1.getCopy(
              b2.getLowerBound(),
              b1.getUpperBound(),
              b2.getLowerBoundType(),
              b1.getUpperBoundType()));
    }
    return Optional.empty();
  }

  private Optional<AbstractGrowingBoundary> createrBoolean(BooleanBoundary b1, BooleanBoundary b2) {
    if ((b1.getLowerBound() == b1.getUpperBound())
        && (b2.getLowerBound() == b2.getUpperBound())
        && b1.getLowerBound() != b1.getLowerBound()) {
      return Optional.of(new BooleanBoundary(""));
    } else {
      return Optional.empty();
    }
  }
}
