package de.unikoblenz.fgbks.base.utils.boundary.bicreater;

import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_NOT_IN_CONTACT;

import de.unikoblenz.fgbks.base.utils.boundary.impl.StringBoundary;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

/**
 * The intersection creator implementation of {@link BoundaryBiCreaterType#INTERSECTION} for {@link
 * StringBoundary}. <br> The result is a new boundary with intersection of both boundaries, or an
 * empty optional, if the creation is not possible. <br> In: {@code b1: "a" b2: "a","b"} <br> Out:
 * {@code "a"} <br>
 */
public class BiCreaterStringIntersection extends AbstractBoundaryBiCreater<StringBoundary> {

  private static final BiCreaterStringIntersection instance = new BiCreaterStringIntersection();

  private BiCreaterStringIntersection() {
    super();
  }

  /**
   * Get the current instance of this class.
   *
   * @return the singleton instance
   */
  public static BiCreaterStringIntersection getInstance() {
    return instance;
  }

  @Override
  public BoundaryBiCreaterType getType() {
    return BoundaryBiCreaterType.INTERSECTION;
  }

  @Override
  public Optional<StringBoundary> create(StringBoundary b1, StringBoundary b2) {
    if (b1.checkWith(IS_NOT_IN_CONTACT, b2)) {
      return Optional.empty();
    }
    if (b1.matchesAny() && b2.matchesAny()) {
      return Optional.of(StringBoundary.getBuilder().matchesAny(true).build());
    }
    if (b1.matchesAny() && !b2.matchesNoneOfValues()
        || b2.matchesAny() && !b1.matchesNoneOfValues()) {
      return Optional.of(
          StringBoundary.getBuilder()
              .addValues(b1.matchesAny() ? b2.getValues() : b1.getValues())
              .build());
    }
    if (b1.matchesAny() && b2.matchesNoneOfValues()) {
      return Optional.of(
          StringBoundary.getBuilder().matchesNoneOfValues(true).addValues(b2.getValues()).build());
    }
    if (b2.matchesAny() && b1.matchesNoneOfValues()) {
      return Optional.of(
          StringBoundary.getBuilder().matchesNoneOfValues(true).addValues(b1.getValues()).build());
    }
    if (b1.matchesNoneOfValues() && b2.matchesNoneOfValues()) {
      Set<String> intersectionStrings = new HashSet<>();
      intersectionStrings.addAll(Arrays.asList(b1.getValues()));
      intersectionStrings.addAll(Arrays.asList(b2.getValues()));
      if (intersectionStrings.size() == 0) {
        return Optional.empty();
      }
      return Optional.of(
          StringBoundary.getBuilder()
              .matchesNoneOfValues(true)
              .addValues(b1.getValues())
              .addValues(b2.getValues())
              .build());
    }
    if (!b1.matchesNoneOfValues() && !b2.matchesNoneOfValues()) {
      Set<String> intersectionStrings = new HashSet<>();
      String[] b1v = b1.getValues();
      int[] b2Hashes = b2.getValuesHashes();
      for (String s1 : b1v) {
        int h1 = Objects.hashCode(s1);
        for (int h2 : b2Hashes) {
          if (h1 == h2) {
            intersectionStrings.add(s1);
            break;
          }
        }
      }
      return Optional.of(StringBoundary.getBuilder().addValues(intersectionStrings).build());
    }
    if (b1.matchesNoneOfValues()) {
      return Optional.of(
          StringBoundary.getBuilder().addValues(getIntersectionStrings(b2, b1)).build());
    }
    if (b2.matchesNoneOfValues()) {
      return Optional.of(
          StringBoundary.getBuilder().addValues(getIntersectionStrings(b1, b2)).build());
    }
    return Optional.empty();
  }

  private Set<String> getIntersectionStrings(StringBoundary b1, StringBoundary b2) {
    Set<String> intersectionStrings = new HashSet<>(Arrays.asList(b1.getValues()));
    int[] b2Hashes = b2.getValuesHashes();
    for (Iterator<String> it = intersectionStrings.iterator(); it.hasNext(); ) {
      String element = it.next();
      int h1 = Objects.hashCode(element);
      for (int h2 : b2Hashes) {
        if (h1 == h2) {
          it.remove();
          break;
        }
      }
    }
    return intersectionStrings;
  }
}
