package de.unikoblenz.fgbks.base.utils.boundary.checker;

import de.unikoblenz.fgbks.base.utils.boundary.impl.StringBoundary;

/**
 * The subsumes checker implementation of {@link BoundaryCheckType#SUBSUMES} for {@link
 * StringBoundary}. <br> The check returns true, if the first boundary completely subsumes (not
 * equal) the other boundary.
 * <br>
 * E.g: <br>
 *
 * <ul>
 *   <li>b1 = "a","b"; b2 = "a" -{@literal >} true
 *   <li>b1 = "a"; b2 = "b" -{@literal >} false
 * </ul>
 */
public class CheckStringSubsumes extends AbstractBoundaryCheck<StringBoundary> {

  private static CheckStringSubsumes instance = new CheckStringSubsumes();

  private CheckStringSubsumes() {
    super();
  }

  /**
   * Get the current instance of this class.
   *
   * @return the singleton instance
   */
  public static CheckStringSubsumes getInstance() {
    return instance;
  }

  @Override
  public BoundaryCheckType getType() {
    return BoundaryCheckType.IS_NOT_IN_CONTACT;
  }

  @Override
  public boolean check(StringBoundary b1, StringBoundary b2) {
    return checkSubsumes(b1, b2);
  }

  /**
   * see {@link CheckStringSubsumes}
   */
  public static boolean checkSubsumes(StringBoundary b1, StringBoundary b2) {
    if (b1.matchesAny() && !b2.matchesAny()) {
      return true;
    }
    if (b2.matchesAny()) {
      return false;
    }
    int[] b1h = b1.getValuesHashes();
    int[] b2h = b2.getValuesHashes();
    // not() and not not()
    if (b1.matchesNoneOfValues() && !b2.matchesNoneOfValues()) {
      for (int i = 0; i < b1h.length; i++) {
        for (int u = 0; u < b2h.length; u++) {
          if (b1h[i] == b2h[u]) {
            return false;
          }
        }
      }
      return true;
    }
    // both not()
    if (b1.matchesNoneOfValues() && b2.matchesNoneOfValues()) {
      return checkX(b1h, b2h, b1h.length >= b2h.length);
    }
    // both normal
    return checkX(b2h, b1h, b1h.length <= b2h.length);
  }

  private static boolean checkX(int[] b1h, int[] b2h, boolean b) {
    if (b) {
      return false;
    }
    for (int i : b1h) {
      boolean found = false;
      for (int u : b2h) {
        found |= i == u;
        if (found) {
          break;
        }
      }
      if (!found) {
        return false;
      }
    }
    return true;
  }
}
