package de.unikoblenz.fgbks.base.value;

import static java.util.Objects.nonNull;
import static org.apache.commons.lang3.Validate.notNull;

import java.io.Serializable;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import javax.xml.bind.annotation.XmlTransient;

@XmlTransient
public abstract class AbstractNumberValueObject<
        T extends AbstractNumberValueObject<T, V>, V extends Number>
    implements Serializable, Comparable<AbstractNumberValueObject<T, V>>, SimpleValueObject {

  public static final String VALUE = "value";

  private V value;

  protected AbstractNumberValueObject() {
    // required for proxying
  }

  public AbstractNumberValueObject(V initialValue) {
    value = validateAndNormalize(initialValue);
  }

  public V getValue() {
    return value;
  }

  @Override
  public String toString() {
    return getValue().toString();
  }

  @Override
  public int hashCode() {
    return getValue().hashCode();
  }

  protected abstract V getMinValue();

  protected abstract V getMaxValue();

  protected V validateAndNormalize(V newValue) {
    notNull(newValue, "Value should not be null");
    if (nonNull(getMinValue()) && compareTo(getMinValue(), newValue) > 0) {
      throw new IllegalArgumentException("Minimum value is " + getMinValue().toString());
    }
    if (nonNull(getMaxValue()) && compareTo(getMaxValue(), newValue) < 0) {
      throw new IllegalArgumentException("Maximum value is " + getMaxValue().toString());
    }
    return newValue;
  }

  public T add(T addValue) {
    return add(addValue.getValue());
  }

  public abstract T add(V addValue);

  public T subtract(T subtractValue) {
    return subtract(subtractValue.getValue());
  }

  public abstract T subtract(V subtractValue);

  public T multiply(T multiplyValue) {
    return multiply(multiplyValue.getValue());
  }

  public abstract T multiply(V multiplyValue);

  public T divide(T divideValue) {
    return divide(divideValue.getValue());
  }

  public abstract T divide(V divideValue);

  public boolean greaterThan(T checkValue) {
    return greaterThan(checkValue.getValue());
  }

  public boolean greaterThan(V checkValue) {
    return compareTo(checkValue) > 0;
  }

  public boolean greaterEqualThan(T checkValue) {
    return greaterEqualThan(checkValue.getValue());
  }

  public boolean greaterEqualThan(V checkValue) {
    return compareTo(checkValue) >= 0;
  }

  public boolean lessEqualThan(T checkValue) {
    return lessEqualThan(checkValue.getValue());
  }

  public boolean lessEqualThan(V checkValue) {
    return compareTo(checkValue) <= 0;
  }

  public boolean lessThan(T checkValue) {
    return lessThan(checkValue.getValue());
  }

  public boolean lessThan(V checkValue) {
    return compareTo(checkValue) < 0;
  }

  public boolean equalTo(T checkValue) {
    return equalTo(checkValue.getValue());
  }

  public boolean equalTo(V checkValue) {
    return compareTo(checkValue) == 0;
  }

  protected T newInstance(V initialValue) {
    try {
      Type genericSuperclass = this.getClass().getGenericSuperclass();
      Class aClass = (Class) ((ParameterizedType) genericSuperclass).getActualTypeArguments()[0];
      Class valueClass = getValue().getClass();
      return (T) aClass.getDeclaredConstructor(valueClass).newInstance(initialValue);
    } catch (Exception e) {
      throw new IllegalStateException(e);
    }
  }

  @Override
  public boolean equals(Object object) {
    if (this == object) {
      return true;
    }
    if (object == null || !getClass().isAssignableFrom(object.getClass())) {
      return false;
    }
    AbstractNumberValueObject<T, V> valueObject = (AbstractNumberValueObject<T, V>) object;
    return valueObject.getValue().equals(getValue());
  }

  @Override
  public int compareTo(AbstractNumberValueObject<T, V> object) {
    return compareTo(object.getValue());
  }

  protected int compareTo(V object) {
    return compareTo(getValue(), object);
  }

  protected int compareTo(V object1, V object2) {
    if (Comparable.class.isAssignableFrom(object1.getClass())
        && Comparable.class.isAssignableFrom(object2.getClass())) {
      return ((Comparable) object1).compareTo(object2);
    }
    throw new UnsupportedOperationException("value is not comparable");
  }
}
